/*
 * This is a free program sample that may be reproduced in any form.
 * The author's information should be retained to preserve its identity.
 *
 * Date written: January 1, 2002
 * Written by: Peraphon Sophatsathit
 * Department of Mathematics, Faculty of Science, Chulalongkorn University.
 * email: Peraphon.S@chula.ac.th, http://pioneer.netserv.chula.ac.th/~sperapho
 *
 * Distributed Operating Systems (2301462) classnote.
 * Description: This sample server module illustrates remote host
 *		communication over standard TCP/IP connection.
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<errno.h>

/*
 * porting from BSD to SVR4
 */
#ifdef	__USE_BSD
#  include	<machine/param.h>
#endif

#include	<sys/types.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<signal.h>
#include	<unistd.h>
#include	<time.h>

/*
 * prototypes
 */
unsigned int	alarm(unsigned int);
void		(*signal(int, void (*disp)(int))) (int);
void		handler(int);
void		t_out(int);
void		bzero(void *, size_t);
int		socket(int, int, int);
int		bind(int, const struct sockaddr *, socklen_t);
int		listen(int, int);
int		accept(int, struct sockaddr *, socklen_t *);

int		proc_loop(int);
int		driver(int, int, int, char *);
int		str_echo(int);
void		clear_buff(char *, int);

/*
 * globals
 */
#define		Small		20
#define		TRUE		1
#define		FALSE		0
#define		LISTENQ		1024
#define		SERV_PORT	9877
#define		Null_char	'\0'

/*
 * error return code
 */
#define		Normal		0
#define		Err_socket	1
#define		Err_bind	2
#define		Err_listen	3
#define		Err_accept	4
#define		Err_connect	5
#define		Err_write	6
#define		Err_read	7
#define		Err_IP		10
#define		Err_fork	88
#define		Err_usage	99

int		flag = FALSE;

/*
 * The purpose of signal calls employed in this program is to prevent
 * runaway processing.  The user may terminate (kill) the process any time
 * via user command (ctrl C) or timer.  The latter can be set to any
 * positive integer ranging from 0 to N (N is recommneded to be small to
 * have any effect).  All signals may appear to have no effect if control
 * is suspended by '(blocking) read'.  In which case, one must send a
 * message by typing from keyboard to get out of 'read' wait.
 * Note that in order for the signals to have an immediate effect,
 * non-blocking read must be set along with extra precaution to handle
 * any 'non-blocking' timing and synchronization idiosyncracies.
 */
int
main(int ac, char **av)
{
	int		rt_code;
	unsigned int	sec;

	signal(SIGINT, handler);
	signal(SIGQUIT, handler);
	signal(SIGALRM, t_out);

	switch (ac)
	{
		case 2:
			sec = atoi(av[1]);
			if (sec > 0)
				alarm(sec);
			break;
		default:
			printf("\nUsage: %s   wait_sec\n\n", av[0]);
			printf("Example: %s   0  (no timer is set)\n", av[0]);
			printf("Example: %s   3  (3 seconds timeout)\n\n", av[0]);
			printf("Wait time for 0 second is recommended.  Any other values\n");
			printf("can be used as a precaution to prevent the process from\n");
			printf("running away, but will cause an abnormal termination.\n");
			printf("However, too long a wait will have no effect if the process\n");
			printf("has already terminated.\n");
			return Err_usage;
	}
	rt_code = proc_loop((int)sec);
	if (rt_code > Normal || flag == TRUE)
		printf("Abnornal termination of RPC loop\n");
	fflush(stdout);
	fflush(stderr);
	return Normal;
}

/*
 * set up standard TCP/IP connection
 */
int
proc_loop(int num)
{
	int			counter = 0;
	int			listenfd, connfd;
	char			buf[BUFSIZ];
	time_t			ticks;
	socklen_t		clilen;
	struct	sockaddr_in	servaddr, cliaddr;

	/*
	 * open a socket to accept incoming request from client(s)
	 */
	if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		return Err_socket;
	}
	bzero(&servaddr, sizeof(servaddr));
	servaddr.sin_family	 = AF_INET;
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servaddr.sin_port 	 = htons(SERV_PORT);

	if (bind(listenfd, (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0)
	{
		return Err_bind;
	}
	if (listen(listenfd, LISTENQ) < 0)
	{
		return Err_listen;
	}
	clilen = sizeof(cliaddr);
	if ((connfd = accept(listenfd, (struct sockaddr *)&cliaddr, &clilen)) < 0)
	{
		return Err_accept;
	}
	ticks  = time(NULL);
	snprintf(buf, sizeof(buf), "%.24s\n", ctime(&ticks));
	if (driver(counter, listenfd, connfd, buf) > 0)
	{
		perror("fork and exec failed");
	}
	close(connfd);
	return Normal;
}

/*
 * The driver function spawns a child process to start a TCP socket to
 * communicate with the client process.
 */
int
driver(int count, int listenfd, int connfd, char *sm)
{
	int	pid = 0;
	int	rt  = 0;

	pid = fork();
	if (pid == 0)
	{
		printf("begin child process\n");
		close(listenfd);
		rt = str_echo(connfd);
	}
	else if (pid > 0)
	{
		printf("parent: spawn succeeded!\n");
	}
	else
	{
		printf("fork failed: parent exiting...\n");
		rt = Err_fork;
	}
	return rt;
}

/*
 * read loop: first send prompt string to client and enter read/receive
 * message loop.  The process terminates when ctrl-D is received or
 * interrupts from pending signals.
 */
int
str_echo(int sockfd)
{
	int	n;
	char	line[BUFSIZ];

	strcpy(line, "begin typing message, ctrl-D to quit\n");
	n = strlen(line);
	write(sockfd, line, n);
	clear_buff(line, n);
	for (; flag == FALSE; )
	{
		if ((n = read(sockfd, line, BUFSIZ)) == 0)
		{
			printf("connection closed by other end\n");
			break;
		}
		write(sockfd, line, n);
		fprintf(stdout, "echo> %s", line);
		clear_buff(line, n);
	}
	return Normal;
}

/*
 * clear R/W buffer to null
 */
void
clear_buff(char *line, int n)
{
	register int	i;

	for (i = 0; i < n; i++)
		line[i] = Null_char;
	return;
}

/*
 * time out by alarm clock
 */
void
t_out(int sig)
{
	signal(SIGALRM, t_out);
	flag = TRUE;
	printf("timeout by ALARM signal\n");
	return;
}

/*
 * interrupt and kill signals
 */
void
handler(int sig)
{
	signal(SIGINT, handler);
	signal(SIGQUIT, handler);
	flag = TRUE;
	printf("receiving INT/QUIT signal\n");
	return;
}
